package travel.iko.sso.domain.security.service;

import lombok.RequiredArgsConstructor;
import org.apache.commons.lang3.SerializationUtils;
import org.springframework.dao.DataRetrievalFailureException;
import org.springframework.data.mongodb.core.query.Criteria;
import org.springframework.data.mongodb.core.query.Query;
import org.springframework.security.oauth2.core.AuthorizationGrantType;
import org.springframework.security.oauth2.core.OAuth2AccessToken;
import org.springframework.security.oauth2.core.OAuth2AuthorizationCode;
import org.springframework.security.oauth2.core.OAuth2RefreshToken;
import org.springframework.security.oauth2.core.OAuth2TokenType;
import org.springframework.security.oauth2.core.endpoint.OAuth2ParameterNames;
import org.springframework.security.oauth2.core.oidc.OidcIdToken;
import org.springframework.security.oauth2.server.authorization.OAuth2Authorization;
import org.springframework.security.oauth2.server.authorization.OAuth2AuthorizationService;
import org.springframework.security.oauth2.server.authorization.client.RegisteredClient;
import org.springframework.security.oauth2.server.authorization.client.RegisteredClientRepository;
import org.springframework.stereotype.Service;
import org.springframework.util.StringUtils;
import travel.iko.sso.domain.security.api.OAuth2AccessTokenType;
import travel.iko.sso.domain.security.api.OAuth2AuthorizationGrantType;
import travel.iko.sso.domain.security.repository.AuthorizationRepository;

import javax.validation.constraints.NotNull;
import java.nio.charset.StandardCharsets;
import java.time.Instant;
import java.util.Map;
import java.util.stream.Collectors;

import static org.springframework.data.mongodb.core.query.Criteria.where;
import static org.springframework.data.mongodb.core.query.Query.query;

/**
 * Created on: 9/14/21.
 *
 * @author Bjorn Harvold Responsibility:
 */
@RequiredArgsConstructor
@Service("oAuth2AuthorizationService")
public class MongoOAuth2AuthorizationService implements OAuth2AuthorizationService {
    private static final char CLASSPATH_DIVIDER = '.';
    private static final char MONGO_FRIENDLY_DIVIDER = '-';

    private final AuthorizationRepository authorizationRepository;
    private final RegisteredClientRepository registeredClientRepository;

    @Override
    public void save(org.springframework.security.oauth2.server.authorization.OAuth2Authorization authorization) {
        final String state = authorization.getAttribute(OAuth2ParameterNames.STATE);
        final OAuth2Authorization.Token<OAuth2AuthorizationCode> authorizationCode = authorization.getToken(OAuth2AuthorizationCode.class);

        if (authorizationCode == null) {
            throw new DataRetrievalFailureException("The authorizationCode is missing.");
        }

        final OAuth2AuthorizationCode token = authorizationCode.getToken();
        final byte[] authorizationCodeValue = token.getTokenValue().getBytes(StandardCharsets.UTF_8);
        final Instant tokenIssuedAt = token.getIssuedAt();
        final Instant tokenExpiresAt = token.getExpiresAt();
        final OAuth2Authorization.Token<OAuth2AccessToken> accessToken = authorization.getAccessToken();
        byte[] accessTokenValue = null;
        OAuth2AccessTokenType tokenType = null;
        String scopes = null;
        Instant accessTokenIssuedAt = null;
        Instant accessTokenExpiresAt = null;

        if (accessToken != null) {
            accessTokenValue = accessToken.getToken().getTokenValue().getBytes(StandardCharsets.UTF_8);
            accessTokenIssuedAt = accessToken.getToken().getIssuedAt();
            accessTokenExpiresAt = accessToken.getToken().getExpiresAt();

            if (OAuth2AccessToken.TokenType.BEARER.getValue().equalsIgnoreCase(accessToken.getToken().getTokenType().getValue())) {
                tokenType = OAuth2AccessTokenType.BEARER;
            }

            scopes = accessToken.getToken().getScopes() != null && accessToken.getToken().getScopes().size() > 0 ? StringUtils.collectionToDelimitedString(accessToken.getToken().getScopes(), ",") : null;
        }

        final OAuth2Authorization.Token<OidcIdToken> oidcIdToken = authorization.getToken(OidcIdToken.class);

        byte[] oidcIdTokenValue = null;
        Instant oidcIdIssuedAt = null;
        Instant oidcIdExpiresAt = null;
        
        if (oidcIdToken != null) {
            oidcIdTokenValue = oidcIdToken.getToken().getTokenValue().getBytes(StandardCharsets.UTF_8);
            oidcIdIssuedAt = oidcIdToken.getToken().getIssuedAt();
            oidcIdExpiresAt = oidcIdToken.getToken().getExpiresAt();
        }

        final OAuth2Authorization.Token<OAuth2RefreshToken> refreshToken = authorization.getRefreshToken();

        byte[] refreshTokenValue = null;
        Instant refreshTokenIssuedAt = null;
        Instant refreshTokenExpiresAt = null;
        
        if (refreshToken != null) {
            refreshTokenValue = refreshToken.getToken().getTokenValue().getBytes(StandardCharsets.UTF_8);
            refreshTokenIssuedAt = refreshToken.getToken().getIssuedAt();
            refreshTokenExpiresAt = refreshToken.getToken().getExpiresAt();
        }

        final byte[] serializedObject = SerializationUtils.serialize(authorization);

        travel.iko.sso.domain.security.document.OAuth2Authorization toSave = new travel.iko.sso.domain.security.document.OAuth2Authorization(
                authorization.getId(),
                authorization.getRegisteredClientId(),
                authorization.getPrincipalName(),
                resolveOauth2AuthorizationGrantType(authorization.getAuthorizationGrantType()),
                state,
                authorizationCodeValue,
                tokenIssuedAt,
                tokenExpiresAt,
                accessTokenValue,
                accessTokenIssuedAt,
                accessTokenExpiresAt,
                tokenType,
                scopes,
                oidcIdTokenValue,
                oidcIdIssuedAt,
                oidcIdExpiresAt,
                refreshTokenValue,
                refreshTokenIssuedAt,
                refreshTokenExpiresAt,
                serializedObject
        );

        this.authorizationRepository.save(toSave);
    }

    @Override
    public void remove(org.springframework.security.oauth2.server.authorization.OAuth2Authorization authorization) {
        this.authorizationRepository.deleteById(authorization.getId());
    }

    @Override
    public OAuth2Authorization findById(String id) {
        final travel.iko.sso.domain.security.document.OAuth2Authorization authorization = this.authorizationRepository.findById(id).orElse(null);

        return this.buildOAuth2Authorization(authorization);
    }

    @Override
    public OAuth2Authorization findByToken(String token, OAuth2TokenType tokenType) {

        Criteria criteria = null;
        if (tokenType == null) {
            criteria = new Criteria().orOperator(
                    where("state").is(token),
                    where("authorizationCodeValue").is(token.getBytes(StandardCharsets.UTF_8)),
                    where("accessTokenValue").is(token.getBytes(StandardCharsets.UTF_8)),
                    where("refreshTokenValue").is(token.getBytes(StandardCharsets.UTF_8))
            );
        } else if (OAuth2ParameterNames.STATE.equals(tokenType.getValue())) {
            criteria = where("state").is(token);
        } else if (OAuth2ParameterNames.CODE.equals(tokenType.getValue())) {
            criteria = where("authorizationCodeValue").is(token.getBytes(StandardCharsets.UTF_8));
        } else if (OAuth2TokenType.ACCESS_TOKEN.equals(tokenType)) {
            criteria = where("accessTokenValue").is(token.getBytes(StandardCharsets.UTF_8));
        } else if (OAuth2TokenType.REFRESH_TOKEN.equals(tokenType)) {
            criteria = where("refreshTokenValue").is(token.getBytes(StandardCharsets.UTF_8));
        }

        if (criteria == null) {
            return null;
        }

        Query query = query(criteria);

        final travel.iko.sso.domain.security.document.OAuth2Authorization authorization = this.authorizationRepository.findByQuery(query).orElse(null);
        return this.buildOAuth2Authorization(authorization);
    }

    private OAuth2Authorization buildOAuth2Authorization(@NotNull final travel.iko.sso.domain.security.document.OAuth2Authorization authorization) {
        if (authorization == null) {
            return null;
        }

        final String registeredClientId = authorization.getRegisteredClientId();

        final RegisteredClient registeredClient = this.registeredClientRepository.findById(registeredClientId);
        if (registeredClient == null) {
            throw new DataRetrievalFailureException(
                    "The RegisteredClient with id '" + registeredClientId + "' was not found in the RegisteredClientRepository.");
        }

        return SerializationUtils.deserialize(authorization.getSerializedObject());

//        final OAuth2Authorization.Builder builder = OAuth2Authorization.withRegisteredClient(registeredClient);
//        final String id = authorization.getId();
//        final String principalName = authorization.getPrincipalName();
//        final OAuth2AuthorizationGrantType authorizationGrantType = authorization.getAuthorizationGrantType();
//        final Map<String, Object> attributes = replaceMapKeyCharacter(authorization.getAttributes(), MONGO_FRIENDLY_DIVIDER, CLASSPATH_DIVIDER);
//
//        builder.id(id)
//                .principalName(principalName)
//                .authorizationGrantType(resolveAuthorizationGrantType(authorizationGrantType))
//                .attributes((attrs) -> attrs.putAll(attributes));
//
//        final String state = authorization.getState();
//        if (StringUtils.hasText(state)) {
//            builder.attribute(OAuth2ParameterNames.STATE, state);
//        }
//
//        String tokenValue;
//        Instant tokenIssuedAt;
//        Instant tokenExpiresAt;
//        byte[] authorizationCodeValue = authorization.getAuthorizationCodeValue();
//
//        if (authorizationCodeValue != null) {
//            tokenValue = new String(authorizationCodeValue, StandardCharsets.UTF_8);
//            tokenIssuedAt = authorization.getAuthorizationCodeIssuedAt();
//            tokenExpiresAt = authorization.getAuthorizationCodeExpiresAt();
//            Map<String, Object> authorizationCodeMetadata = replaceMapKeyCharacter(authorization.getAuthorizationCodeMetadata(), MONGO_FRIENDLY_DIVIDER, CLASSPATH_DIVIDER);
//
//            OAuth2AuthorizationCode authorizationCode = new OAuth2AuthorizationCode(
//                    tokenValue, tokenIssuedAt, tokenExpiresAt);
//            builder.token(authorizationCode, (metadata) -> metadata.putAll(authorizationCodeMetadata));
//        }
//
//        byte[] accessTokenValue = authorization.getAccessTokenValue();
//        if (accessTokenValue != null) {
//            tokenValue = new String(accessTokenValue, StandardCharsets.UTF_8);
//            tokenIssuedAt = authorization.getAccessTokenIssuedAt();
//            tokenExpiresAt = authorization.getAccessTokenExpiresAt();
//            Map<String, Object> accessTokenMetadata = replaceMapKeyCharacter(authorization.getAccessTokenMetadata(), MONGO_FRIENDLY_DIVIDER, CLASSPATH_DIVIDER);
//            OAuth2AccessToken.TokenType tokenType = null;
//            if (OAuth2AccessToken.TokenType.BEARER.getValue().equalsIgnoreCase(authorization.getAccessTokenType().getValue())) {
//                tokenType = OAuth2AccessToken.TokenType.BEARER;
//            }
//
//            Set<String> scopes = Collections.emptySet();
//            String accessTokenScopes = authorization.getAccessTokenScopes();
//            if (accessTokenScopes != null) {
//                scopes = StringUtils.commaDelimitedListToSet(accessTokenScopes);
//            }
//            OAuth2AccessToken accessToken = new OAuth2AccessToken(tokenType, tokenValue, tokenIssuedAt, tokenExpiresAt, scopes);
//            builder.token(accessToken, (metadata) -> metadata.putAll(accessTokenMetadata));
//        }
//
//        byte[] oidcIdTokenValue = authorization.getOidcIdTokenValue();
//        if (oidcIdTokenValue != null) {
//            tokenValue = new String(oidcIdTokenValue, StandardCharsets.UTF_8);
//            tokenIssuedAt = authorization.getOidcIdTokenIssuedAt();
//            tokenExpiresAt = authorization.getOidcIdTokenExpiresAt();
//            Map<String, Object> oidcTokenMetadata = replaceMapKeyCharacter(authorization.getOidcIdTokenMetadata(), MONGO_FRIENDLY_DIVIDER, CLASSPATH_DIVIDER);
//
//            OidcIdToken oidcToken = new OidcIdToken(
//                    tokenValue, tokenIssuedAt, tokenExpiresAt, (Map<String, Object>) oidcTokenMetadata.get(OAuth2Authorization.Token.CLAIMS_METADATA_NAME));
//            builder.token(oidcToken, (metadata) -> metadata.putAll(oidcTokenMetadata));
//        }
//
//        byte[] refreshTokenValue = authorization.getRefreshTokenValue();
//        if (refreshTokenValue != null) {
//            tokenValue = new String(refreshTokenValue, StandardCharsets.UTF_8);
//            tokenIssuedAt = authorization.getRefreshTokenIssuedAt();
//            tokenExpiresAt = authorization.getRefreshTokenExpiresAt();
//
//            Map<String, Object> refreshTokenMetadata = replaceMapKeyCharacter(authorization.getRefreshTokenMetadata(), MONGO_FRIENDLY_DIVIDER, CLASSPATH_DIVIDER);
//
//            OAuth2RefreshToken refreshToken = new OAuth2RefreshToken(
//                    tokenValue, tokenIssuedAt, tokenExpiresAt);
//            builder.token(refreshToken, (metadata) -> metadata.putAll(refreshTokenMetadata));
//        }
//        return builder.build();

    }

    private AuthorizationGrantType resolveAuthorizationGrantType(OAuth2AuthorizationGrantType authorizationGrantType) {
        if (AuthorizationGrantType.AUTHORIZATION_CODE.getValue().equals(authorizationGrantType.getValue())) {
            return AuthorizationGrantType.AUTHORIZATION_CODE;
        } else if (AuthorizationGrantType.CLIENT_CREDENTIALS.getValue().equals(authorizationGrantType.getValue())) {
            return AuthorizationGrantType.CLIENT_CREDENTIALS;
        } else if (AuthorizationGrantType.REFRESH_TOKEN.getValue().equals(authorizationGrantType.getValue())) {
            return AuthorizationGrantType.REFRESH_TOKEN;
        }
        return new AuthorizationGrantType(authorizationGrantType.getValue());        // Custom authorization grant type
    }

    private OAuth2AuthorizationGrantType resolveOauth2AuthorizationGrantType(AuthorizationGrantType authorizationGrantType) {
        if (OAuth2AuthorizationGrantType.AUTHORIZATION_CODE.getValue().equals(authorizationGrantType.getValue())) {
            return OAuth2AuthorizationGrantType.AUTHORIZATION_CODE;
        } else if (OAuth2AuthorizationGrantType.CLIENT_CREDENTIALS.getValue().equals(authorizationGrantType.getValue())) {
            return OAuth2AuthorizationGrantType.CLIENT_CREDENTIALS;
        } else if (OAuth2AuthorizationGrantType.REFRESH_TOKEN.getValue().equals(authorizationGrantType.getValue())) {
            return OAuth2AuthorizationGrantType.REFRESH_TOKEN;
        }
        return OAuth2AuthorizationGrantType.PASSWORD;        // FYI this is different than the example so might be an issue
    }

    private Map<String, Object> replaceMapKeyCharacter(Map<String, Object> map, char toReplace, char replaceWith) {
        return map != null && map.size() > 0 ? map.entrySet()
                .stream()
                .collect(Collectors.toMap(entry -> entry.getKey().replace(toReplace, replaceWith), Map.Entry::getValue)) : null;
    }
}
