package travel.iko.sso.domain.security.service;

import lombok.RequiredArgsConstructor;
import org.springframework.security.oauth2.core.AuthorizationGrantType;
import org.springframework.security.oauth2.core.ClientAuthenticationMethod;
import org.springframework.security.oauth2.server.authorization.client.RegisteredClient;
import org.springframework.security.oauth2.server.authorization.client.RegisteredClientRepository;
import org.springframework.security.oauth2.server.authorization.config.ClientSettings;
import org.springframework.security.oauth2.server.authorization.config.TokenSettings;
import org.springframework.stereotype.Component;
import travel.iko.sso.domain.security.api.OAuth2AuthorizationGrantType;
import travel.iko.sso.domain.security.api.OAuth2ClientAuthenticationMethod;
import travel.iko.sso.domain.security.api.OAuth2ClientSettings;
import travel.iko.sso.domain.security.api.OAuth2TokenSettings;
import travel.iko.sso.domain.security.document.OAuth2RegisteredClient;

import java.time.Instant;
import java.util.List;
import java.util.stream.Collectors;

/**
 * Created on: 9/13/21.
 *
 * @author Bjorn Harvold Responsibility:
 */
@RequiredArgsConstructor
@Component
public class MongoRegisteredClientRepository implements RegisteredClientRepository {
    private final travel.iko.sso.domain.security.repository.RegisteredClientRepository repository;

    @Override
    public void save(RegisteredClient registeredClient) {
        Instant clientIdIssuedAt = registeredClient.getClientIdIssuedAt() != null ?
                registeredClient.getClientIdIssuedAt() : Instant.now();

        Instant clientSecretExpiresAt = registeredClient.getClientSecretExpiresAt() != null ?
                registeredClient.getClientSecretExpiresAt() : null;

        final List<OAuth2ClientAuthenticationMethod> clientAuthenticationMethods = registeredClient.getClientAuthenticationMethods().stream()
                .map(MongoRegisteredClientRepository::resolveOauth2ClientAuthenticationMethod)
                .toList();

        final List<OAuth2AuthorizationGrantType> authorizationGrantTypes = registeredClient.getAuthorizationGrantTypes().stream()
                .map(MongoRegisteredClientRepository::resolveOauth2AuthorizationGrantType)
                .toList();

        final OAuth2ClientSettings clientSettings = new OAuth2ClientSettings(registeredClient.getClientSettings().isRequireProofKey(), registeredClient.getClientSettings().isRequireAuthorizationConsent());
        OAuth2TokenSettings tokenSettings = new OAuth2TokenSettings(registeredClient.getTokenSettings().getAccessTokenTimeToLive(), registeredClient.getTokenSettings().isReuseRefreshTokens(), registeredClient.getTokenSettings().getRefreshTokenTimeToLive(), registeredClient.getTokenSettings().getIdTokenSignatureAlgorithm());

        final OAuth2RegisteredClient client = new OAuth2RegisteredClient(
                registeredClient.getId(),
                registeredClient.getClientId(),
                clientIdIssuedAt,
                registeredClient.getClientSecret(),
                clientSecretExpiresAt,
                registeredClient.getClientName(),
                clientAuthenticationMethods,
                authorizationGrantTypes,
                registeredClient.getRedirectUris(),
                registeredClient.getScopes(),
                clientSettings,
                tokenSettings
        );

        this.repository.save(client);
    }

    @Override
    public RegisteredClient findById(String id) {
        final OAuth2RegisteredClient client = this.repository.findById(id).orElse(null);

        return buildRegisteredClient(client);
    }

    @Override
    public RegisteredClient findByClientId(String clientId) {
        final OAuth2RegisteredClient client = this.repository.findByClientId(clientId).orElse(null);

        return buildRegisteredClient(client);
    }

    private RegisteredClient buildRegisteredClient(OAuth2RegisteredClient client) {
        if (client == null) {
            return null;
        }

        final List<ClientAuthenticationMethod> methods = client.getClientAuthenticationMethods().stream()
                .map(MongoRegisteredClientRepository::resolveClientAuthenticationMethod)
                .toList();

        final List<AuthorizationGrantType> grantTypes = client.getAuthorizationGrantTypes().stream()
                .map(MongoRegisteredClientRepository::resolveAuthorizationGrantType)
                .toList();

        final ClientSettings clientSettings = ClientSettings.builder()
                .requireProofKey(client.getClientSettings().isRequireProofKey())
                .requireAuthorizationConsent(client.getClientSettings().isRequireAuthorizationConsent())
                .build();

        final TokenSettings tokenSettings = TokenSettings.builder()
                .accessTokenTimeToLive(client.getTokenSettings().getAccessTokenTimeToLive())
                .reuseRefreshTokens(client.getTokenSettings().isReuseRefreshTokens())
                .refreshTokenTimeToLive(client.getTokenSettings().getRefreshTokenTimeToLive())
                .idTokenSignatureAlgorithm(client.getTokenSettings().getIdTokenSignatureAlgorithm())
                .build();

        return RegisteredClient.withId(client.getId())
                .clientId(client.getClientId())
                .clientIdIssuedAt(client.getClientIdIssuedAt())
                .clientSecret(client.getClientSecret())
                .clientSecretExpiresAt(client.getClientSecretExpiresAt())
                .clientName(client.getClientName())
                .clientAuthenticationMethods(clientMethods ->
                        clientMethods.addAll(methods)
                )
                .authorizationGrantTypes(authorizationTypes ->
                        authorizationTypes.addAll(grantTypes)
                )
                .redirectUris(redirectUris -> redirectUris.addAll(client.getRedirectUris()))
                .scopes(scopes -> scopes.addAll(client.getScopes()))
                .clientSettings(clientSettings)
                .tokenSettings(tokenSettings)
                .build();
    }

    private static ClientAuthenticationMethod resolveClientAuthenticationMethod(OAuth2ClientAuthenticationMethod clientAuthenticationMethod) {
        if (ClientAuthenticationMethod.CLIENT_SECRET_BASIC.getValue().equals(clientAuthenticationMethod.getValue())) {
            return ClientAuthenticationMethod.CLIENT_SECRET_BASIC;
        } else if (ClientAuthenticationMethod.CLIENT_SECRET_POST.getValue().equals(clientAuthenticationMethod.getValue())) {
            return ClientAuthenticationMethod.CLIENT_SECRET_POST;
        } else if (ClientAuthenticationMethod.NONE.getValue().equals(clientAuthenticationMethod.getValue())) {
            return ClientAuthenticationMethod.NONE;
        }
        return new ClientAuthenticationMethod(clientAuthenticationMethod.getValue());        // Custom client authentication method
    }

    private static OAuth2ClientAuthenticationMethod resolveOauth2ClientAuthenticationMethod(ClientAuthenticationMethod clientAuthenticationMethod) {
        if (OAuth2ClientAuthenticationMethod.CLIENT_SECRET_BASIC.getValue().equals(clientAuthenticationMethod.getValue())) {
            return OAuth2ClientAuthenticationMethod.CLIENT_SECRET_BASIC;
        } else if (OAuth2ClientAuthenticationMethod.CLIENT_SECRET_POST.getValue().equals(clientAuthenticationMethod.getValue())) {
            return OAuth2ClientAuthenticationMethod.CLIENT_SECRET_POST;
        } else if (OAuth2ClientAuthenticationMethod.NONE.getValue().equals(clientAuthenticationMethod.getValue())) {
            return OAuth2ClientAuthenticationMethod.NONE;
        }
        return OAuth2ClientAuthenticationMethod.NONE;        // Custom client authentication method
    }

    private static AuthorizationGrantType resolveAuthorizationGrantType(OAuth2AuthorizationGrantType authorizationGrantType) {
        if (AuthorizationGrantType.AUTHORIZATION_CODE.getValue().equals(authorizationGrantType.getValue())) {
            return AuthorizationGrantType.AUTHORIZATION_CODE;
        } else if (AuthorizationGrantType.CLIENT_CREDENTIALS.getValue().equals(authorizationGrantType.getValue())) {
            return AuthorizationGrantType.CLIENT_CREDENTIALS;
        } else if (AuthorizationGrantType.REFRESH_TOKEN.getValue().equals(authorizationGrantType.getValue())) {
            return AuthorizationGrantType.REFRESH_TOKEN;
        }
        return new AuthorizationGrantType(authorizationGrantType.getValue());        // Custom authorization grant type
    }

    private static OAuth2AuthorizationGrantType resolveOauth2AuthorizationGrantType(AuthorizationGrantType authorizationGrantType) {
        if (OAuth2AuthorizationGrantType.AUTHORIZATION_CODE.getValue().equals(authorizationGrantType.getValue())) {
            return OAuth2AuthorizationGrantType.AUTHORIZATION_CODE;
        } else if (OAuth2AuthorizationGrantType.CLIENT_CREDENTIALS.getValue().equals(authorizationGrantType.getValue())) {
            return OAuth2AuthorizationGrantType.CLIENT_CREDENTIALS;
        } else if (OAuth2AuthorizationGrantType.REFRESH_TOKEN.getValue().equals(authorizationGrantType.getValue())) {
            return OAuth2AuthorizationGrantType.REFRESH_TOKEN;
        }
        return OAuth2AuthorizationGrantType.PASSWORD;        // FYI this is different than the example so might be an issue
    }
}
